### Baseline PCA function
baseline_pca_si_fn <- function(data, grp) {
  
  return(SilhouetteBenchmarkPerCTPCA(data = data, grp = grp))
  
  # log.data <- log(data + 1)
  # 
  # si_list <- list()
  # num_pcs = 15
  # pca.data <-
  #     irlba::prcomp_irlba(
  #         x = t(log.data),
  #         n = num_pcs,
  #         center = TRUE,
  #         scale. = FALSE
  #     )$x
  # 
  # require(cluster)
  # # PCA_Corr
  # # cor_cell = cor(t(pca.data))
  # # cor_cell[which(is.na(cor_cell))] = 0
  # # 
  # # dis_cor <- as.matrix(as.dist(1 - cor_cell))
  # # if (any(is.na(dis_cor)) |
  # #     any(is.nan(dis_cor)) | any(is.infinite(dis_cor)))
  # #     return(0.0000002)
  # # 
  # # sil_cor <- silhouette(x = grp, dmatrix = dis_cor)
  # # si_list[["PCA_Corr"]] <- mean(sil_cor[, 3])
  # 
  # #PCA_z_euc
  # # pca.z.data <- t(apply(pca.data, 1, function(x) {
  # #     (x - mean(x))/sd(x)
  # # }))
  # 
  # # Euc
  # dis_euc <- as.matrix(dist(pca.data, method = "euclidean"))
  # 
  # if (any(is.na(dis_euc)) |
  #     any(is.nan(dis_euc)) | any(is.infinite(dis_euc)))
  #     return(0.0000002)
  # 
  # sil_z_euc <- silhouette(x = grp, dmatrix = dis_euc)
  # si_list[["PCA_Euc"]] <- mean(sil_z_euc[, 3])
  # 
  # 
  # return(mean(sil_z_euc[, 3]))
}


baseline_pca_ari_fn <- function(data, grp) {
  log.data <- log(data + 1)
  
  si_list <- list()
  num_pcs = 15
  pca.data <-
    irlba::prcomp_irlba(
      x = t(log.data),
      n = num_pcs,
      center = TRUE,
      scale. = FALSE
    )$x
  
  d <- dist(pca.data, method = "euclidean")
  
  hcl <- stats::hclust(d, method = "average")
  cluster <- cutree(hcl, k = length(unique(grp)))
  
  return(adjustedRandIndex(x = cluster, y = grp))
}


### Filter genes
gene_filt_fn <- function(data) {
  filt.data <- data[apply(data, 1, function(x){
    sum(x>0)>0.05*ncol(data)
  }), , drop = FALSE]
  
  return(filt.data)
}


### Filter genes
lowly_expressed_gene_filt_fn <- function(data) {
  filt.data <- data[apply(data, 1, function(x) {
    sum(x>0)>0.001*ncol(data)
  }), , drop = FALSE]
  
  return(filt.data)
}


######################### Feature Selection ############################################################################


cor_range_pov_min_rank_fn <- function(data, data_name) {
  
  cor_range_out <- getFeatureGene(data)
  cor_range_ranks <- rank(x = -cor_range_out$range, ties.method = "average")
  names(cor_range_ranks) <- names(cor_range_out$range)
  
  pov_out <- getFeatureGene_POV_only(data)
  pov_ranks <- rank(x = -pov_out$zs_ranked, ties.method = "average")
  names(pov_ranks) <- names(pov_out$zs_ranked)
  
  min_ranks <- list()
  for(gene in rownames(data)) {
    min_ranks[[gene]] = min(cor_range_ranks[gene], pov_ranks[gene])
  }
  
  min_ranks <- unlist(min_ranks)
  min_ranks_no_ties <- rank(x = min_ranks, ties.method = "average")
  names(min_ranks_no_ties) <- names(min_ranks)
  
  sorted_min_ranks_no_ties <- sort(min_ranks_no_ties)
  
  return(list("ranks" = sorted_min_ranks_no_ties, "genes" = names(sorted_min_ranks_no_ties)))
}


cor_range_pov_geomean_rank_fn <- function(data, data_name) {
  
  cor_range_out <- getFeatureGene(data)
  cor_range_ranks <- rank(x = -cor_range_out$range, ties.method = "average")
  names(cor_range_ranks) <- names(cor_range_out$range)
  
  pov_out <- getFeatureGene_POV_only(data)
  pov_ranks <- rank(x = -pov_out$zs_ranked, ties.method = "average")
  names(pov_ranks) <- names(pov_out$zs_ranked)
  
  geo_mean_ranks <- list()
  for(gene in rownames(data)) {
    geo_mean_ranks[[gene]] = geoMean(x = c(cor_range_ranks[gene], pov_ranks[gene]), na.rm = T)
  }
  
  geo_mean_ranks <- unlist(geo_mean_ranks)
  geo_mean_ranks_no_ties <- rank(x = geo_mean_ranks, ties.method = "average")
  names(geo_mean_ranks_no_ties) <- names(geo_mean_ranks)
  
  sorted_geo_mean_ranks_no_ties <- sort(geo_mean_ranks_no_ties)
  
  return(list("ranks" = sorted_geo_mean_ranks_no_ties, "genes" = names(sorted_geo_mean_ranks_no_ties)))
}

cor_range_pov_sum_zscores_fn <- function(data, data_name) {
  
  cor_range_out <- getFeatureGene(data)
  cor_range_z_scores <- cor_range_out$zscore
  
  pov_out <- getFeatureGene_POV_only(data)
  pov_z_scores <- pov_out$zs_ranked
  
  z_score_sums <- c()
  for(gene in rownames(data)) {
    z_score_sums[gene] = cor_range_z_scores[gene] + pov_z_scores[gene]
  }
  
  sorted_z_score_sums <- sort(z_score_sums, decreasing = T)
  
  return(list("ranks" = sorted_z_score_sums, "genes" = names(sorted_z_score_sums)))
}


all_5_min_rank_fn <- function(data, data_name) {
  
  cor_range_out <- getFeatureGene(data)
  cor_range_ranks <- rank(x = -cor_range_out$range, ties.method = "average")
  names(cor_range_ranks) <- names(cor_range_out$range)
  
  pov_out <- getFeatureGene_POV_only(data)
  pov_ranks <- rank(x = -pov_out$zs_ranked, ties.method = "average")
  names(pov_ranks) <- names(pov_out$zs_ranked)
  
  hcv_out <- getFeatureGene_HCV(data)
  hcv_ranks <- rank(x = -hcv_out$zs_ranked, ties.method = "average")
  names(hcv_ranks) <- names(hcv_out$zs_ranked)
  
  hcv2_out <- getFeatureGene_HCV2(data)
  hcv2_ranks <- rank(x = -hcv2_out$zs_ranked, ties.method = "average")
  names(hcv2_ranks) <- names(hcv2_out$zs_ranked)
  
  hdg_out <- UseHVG2(data)
  hdg_ranks <- rank(x = -hdg_out$zs_ranked, ties.method = "average")
  names(hdg_ranks) <- names(hdg_out$zs_ranked)
  
  min_ranks <- list()
  for(gene in rownames(data)) {
    min_ranks[[gene]] = min(cor_range_ranks[gene], pov_ranks[gene], hcv_ranks[gene], hcv2_ranks[gene], hdg_ranks[gene])
  }
  
  min_ranks <- unlist(min_ranks)
  min_ranks_no_ties <- rank(x = min_ranks, ties.method = "average")
  names(min_ranks_no_ties) <- names(min_ranks)
  
  sorted_min_ranks_no_ties <- sort(min_ranks_no_ties)
  
  return(list("ranks" = sorted_min_ranks_no_ties, "genes" = names(sorted_min_ranks_no_ties)))
  
}


all_5_geomean_rank_fn <- function(data, data_name) {
  
  cor_range_out <- getFeatureGene(data)
  cor_range_ranks <- rank(x = -cor_range_out$range, ties.method = "average")
  names(cor_range_ranks) <- names(cor_range_out$range)
  
  pov_out <- getFeatureGene_POV_only(data)
  pov_ranks <- rank(x = -pov_out$zs_ranked, ties.method = "average")
  names(pov_ranks) <- names(pov_out$zs_ranked)
  
  hcv_out <- getFeatureGene_HCV(data)
  hcv_ranks <- rank(x = -hcv_out$zs_ranked, ties.method = "average")
  names(hcv_ranks) <- names(hcv_out$zs_ranked)
  
  hcv2_out <- getFeatureGene_HCV2(data)
  hcv2_ranks <- rank(x = -hcv2_out$zs_ranked, ties.method = "average")
  names(hcv2_ranks) <- names(hcv2_out$zs_ranked)
  
  hdg_out <- UseHVG2(data)
  hdg_ranks <- rank(x = -hdg_out$zs_ranked, ties.method = "average")
  names(hdg_ranks) <- names(hdg_out$zs_ranked)
  
  geo_mean_ranks <- list()
  for(gene in rownames(data)) {
    geo_mean_ranks[[gene]] = geoMean(x = c(cor_range_ranks[gene], pov_ranks[gene], hcv_ranks[gene], hcv2_ranks[gene], hdg_ranks[gene]), na.rm = T)
  }
  
  geo_mean_ranks <- unlist(geo_mean_ranks)
  geo_mean_ranks_no_ties <- rank(x = geo_mean_ranks, ties.method = "average")
  names(geo_mean_ranks_no_ties) <- names(geo_mean_ranks)
  
  sorted_geo_mean_ranks_no_ties <- sort(geo_mean_ranks_no_ties)
  
  return(list("ranks" = sorted_geo_mean_ranks_no_ties, "genes" = names(sorted_geo_mean_ranks_no_ties)))
  
}



cor_range_pos_neg_z_fn <- function(data, data_name) {
  
  return(getFeatureGene_cor_zs(data))
}

### Local Cor Range function
cor_range_fn <- function(data, data_name) {
  return(getFeatureGene(data))
}

### Local POV function
pov_fn <- function(data, data_name) {
  return(getFeatureGene_POV_only(data))
}

### Local HCV function
hcv_fn <- function(data, data_name) {
  return(getFeatureGene_HCV(data))
}

### Local HCV2 function
hcv2_fn <- function(data, data_name) {
  return(getFeatureGene_HCV2(data))
}

### M3Drop function
m3drop_fn <- function(data, data_name) {
  # data <- as.matrix(data)
  umi_datasets <- c("CBMC_CITE", "PBMC_CITE", "MALT_CITE", "PBMC_Dropseq_CITE", "VDJ_CITE", "Cell_Hashing", "NSCLC", "PBMC_ECCITE_Seq", "5cl_10x", "3cl_10x", "3cl_dropseq", "3cl_celseq", "5cl_celseq", "FACS_PBMC")
  smart_seq_datasets <- c("Melanoma", "CRC_Cell_Line", "Mouse_Brain", "Mixology", "Melanoma_Cancer", "Breast_WY", "GBM", "ODG")
  if(data_name %in% umi_datasets) {
    return(UseM3D(data = data, method = "DANB"))
  }
  else {
    return(UseM3D(data = data, method = "M3Drop"))
  }
  
}

### Local GiniClust function to measure execution time
gini_fn <- function(data, data_name) {
  out.folder = paste0("Giniclust/")
  
  st <- system.time({
    Gini_res <-
      GiniClust_Fitting(
        data.type = "RNA-seq",
        # ExprM.RawCounts.filter = as.matrix(data),
        ExprM.RawCounts.filter = data,
        out.folder = out.folder,
        exprimentID = "data"
      )
    
    
    # res <- list("p.val" = Gini_res$p.value)
    p.val = Gini_res$`p-value`
    names(p.val) <- Gini_res$Gene
    p.val <- sort(p.val)
    genes <- names(p.val)
    
    # Gini_res$score <- -log10(Gini_res$p.value)
    # rownames(Gini_res) <- Gini_res$Gene
  })
  
  return(list(
    "p.val" = p.val,
    "genes" = genes,
    "time" = st
  ))
}

### GGC function (empty since GGC isn't being implemented anymore)


### HVG2 function
hvg2_fn <- function(data, data_name) {
  return(UseHVG2(data))
}


### Local FindVariableFeatures VST function to measure execution time
hvg_vst_fn <- function(data, data_name) {
    seurat_obj <- CreateSeuratObject(counts = data)
    # seurat_obj <- NormalizeData(object = seurat_obj)
    st <- system.time({
      
    seurat_obj <- FindVariableFeatures(object = seurat_obj, selection.method = "vst")
    })
    hvf.info <- seurat_obj@assays$RNA@meta.features
    hvf.info <- hvf.info[order(hvf.info$vst.variance.standardized, decreasing = T), ]

  
  # res <- list("gene.dispersion" = seurat_var_genes_obj@hvg.info$gene.dispersion, "genes" = rownames(seurat_var_genes_obj@hvg.info))
  
  return(list(
    "vst.variance.standardized" = hvf.info$vst.variance.standardized,
    "genes" = rownames(hvf.info),
    "time" = st
  ))
}

### Local FindVariableFeatures dispersion function to measure execution time
hvg_disp_fn <- function(data, data_name) {
    seurat_obj <- CreateSeuratObject(counts = data)
    # seurat_obj <- NormalizeData(object = seurat_obj)
    st <- system.time({
      
    seurat_obj <- FindVariableFeatures(object = seurat_obj, selection.method = "dispersion")
  })
    hvf.info <- seurat_obj@assays$RNA@meta.features
    hvf.info <- hvf.info[order(hvf.info$mvp.dispersion.scaled, decreasing = T), ]
  
  
  # res <- list("gene.dispersion" = seurat_var_genes_obj@hvg.info$gene.dispersion, "genes" = rownames(seurat_var_genes_obj@hvg.info))
  
  return(list(
    "mvp.dispersion.scaled" = hvf.info$mvp.dispersion.scaled,
    "genes" = rownames(hvf.info),
    "time" = st
  ))
}



### Local PCA function to measure execution time
pca_fn <- function(data, data_name) {
  st <- system.time({
    PCA_Score <- irlba_pca_fs(expr_mat = data)
    
  })
  
  
  return(list(
    "score" = PCA_Score,
    "genes" = names(PCA_Score),
    "time" = st
  ))
}

irlba_pca_fs <- function (expr_mat, pcs = c(2, 3)) 
{
  norm <- expr_mat
  nz_genes <- which(Matrix::rowSums(norm) != 0)
  norm[nz_genes, ] <- log(norm[nz_genes, ] + 1)/log(2)
  gene_names <- rownames(norm)
  norm <- Matrix::Matrix(norm, sparse = TRUE)
  rownames(norm) <- gene_names
  nc <- ncol(norm)
  expression_means <- Matrix::rowMeans(norm)
  expression_vars <- Matrix::rowMeans((norm - expression_means)^2) * 
    (nc/(nc - 1))
  genes_to_keep <- expression_vars > 0
  norm <- norm[genes_to_keep, ]
  expression_means <- expression_means[genes_to_keep]
  expression_vars <- expression_vars[genes_to_keep]
  irlba_pca_res <- irlba::irlba(Matrix::t(norm), nu = 0, center = expression_means, nv = 30,
                                scale = sqrt(expression_vars), right_only = TRUE)$v
  row.names(irlba_pca_res) <- row.names(norm)
  if (length(pcs) > 1) {
    score <- Matrix::rowSums(abs(irlba_pca_res[, pcs]))
  }
  else {
    score <- abs(irlba_pca_res[, pcs])
  }
  names(score) = gene_names[genes_to_keep]
  return(sort(-score))
}


scmarker_fn <- function(data, group) {
  scmarker_out <- UseSCMarker(data = data)
  
  scmarker_si <- SilhouetteBenchmarkPerCTPCA(data = data[scmarker_out$genes, ], grp = group)
  
  list_out <- list()
  list_out[[paste0(length(scmarker_out$genes))]] <- scmarker_si
  
  return(list_out)
}


stepwise_reg_fn <- function(data, group) {
  
  # Remove mito genes
  data <- data[!grepl(pattern = "^MT-", x = rownames(data)), ]
  data <- data[!grepl(pattern = "^MT_", x = rownames(data)), ]
  
  # Remove ERCC genes
  data <- data[!grepl(pattern = "^ERCC-", x = rownames(data)), ]
  data <- data[!grepl(pattern = "^ERCC_", x = rownames(data)), ]
  
  # Remove RPL and RPS genes
  data <- data[!grepl(pattern = "^RPS", x = rownames(data)), ]
  data <- data[!grepl(pattern = "^RPL", x = rownames(data)), ]
  
  # Log transform input data matrix
  log.data <- log(1+data)
  
  # Compute gene-gene correlation matrix
  system.time({
    ggc_original = WGCNA::cor(t(as.matrix(log.data)),method="pearson",nThreads = 16)
  })
  
  # Smooth out diagonal by setting second largest correlation (i.e. largest correlation of a gene with another gene)
  # for(i in 1:ncol(ggc_original)) {
  #     ggc_original[i, i] <- sort(ggc_original[i, ], decreasing = T)[2]
  # }
  
  
  ### Step 1: Choose the top genes based on correlation range
  rangeList <- getCorrelationRange(ggc_original)
  topRangeGenes <- names(rangeList$range)[1:1000]
  
  povList <- getPOV(ggc_original)
  topPOVGenes <- names(povList$POV)[1:1000]
  
  topGenes <- union(topRangeGenes, topPOVGenes)
  
  stepwise_reg_obj <- stepwiseRegressionMatrixFS(ggc_original, topGenes, num_genes = 1000)
  print(stepwise_reg_obj$st)
  
  stepwise_reg_ge_obj <- stepwiseRegressionMatrixGEFS(data, ggc_original, topGenes, num_genes = 1000)
  print(stepwise_reg_ge_obj$st)
  
  list_out <- list()
  list_ge_out <- list()
  num_neighbours <- c("0", "1", "5", "10", "20", "30", "50", "70", "100", "200", "400", "800")
  
  for(n in num_neighbours) {
    list_out[[n]] <- list(num_genes = length(stepwise_reg_obj$neighbour_feature_genes[[n]]), si = SilhouetteBenchmarkPerCTPCA(data = data[stepwise_reg_obj$neighbour_feature_genes[[n]], ], grp = group))
    
    list_ge_out[[n]] <- list(num_genes = length(stepwise_reg_ge_obj$neighbour_feature_genes[[n]]), si = SilhouetteBenchmarkPerCTPCA(data = data[stepwise_reg_ge_obj$neighbour_feature_genes[[n]], ], grp = group))
  }
  
  return(list("SWReg" = list_out, "SWReg_GE" = list_ge_out))
}

stepwise_reg_ggc_fn <- function(data, group) {
  
  # Remove mito genes
  data <- data[!grepl(pattern = "^MT-", x = rownames(data)), ]
  data <- data[!grepl(pattern = "^MT_", x = rownames(data)), ]
  
  # Remove ERCC genes
  data <- data[!grepl(pattern = "^ERCC-", x = rownames(data)), ]
  data <- data[!grepl(pattern = "^ERCC_", x = rownames(data)), ]
  
  # Remove RPL and RPS genes
  data <- data[!grepl(pattern = "^RPS", x = rownames(data)), ]
  data <- data[!grepl(pattern = "^RPL", x = rownames(data)), ]
  
  # Log transform input data matrix
  log.data <- log(1+data)
  
  # Compute gene-gene correlation matrix
  system.time({
    ggc_original = WGCNA::cor(t(as.matrix(log.data)),method="pearson",nThreads = 16)
  })
  
  # Smooth out diagonal by setting second largest correlation (i.e. largest correlation of a gene with another gene)
  # for(i in 1:ncol(ggc_original)) {
  #     ggc_original[i, i] <- sort(ggc_original[i, ], decreasing = T)[2]
  # }
  
  ### Choose the top genes based on correlation range
  rangeList <- getCorrelationRange(ggc_original)
  topRangeGenes <- names(rangeList$range)[1:1000]
  
  # povList <- getPOV(ggc_original)
  # topPOVGenes <- names(povList$POV)[1:1000]
  
  # topGenes <- union(topRangeGenes, topPOVGenes)
  # topGenes = topRangeGenes
  
  # Use HVG to select top genes to start regression
  seurat_obj <- CreateSeuratObject(raw.data = data, normalization.method = "LogNormalize", display.progress = F)
  seurat_var_genes_obj <- FindVariableGenes(seurat_obj, do.plot = F, display.progress = F)
  hvg.info = seurat_var_genes_obj@hvg.info
  hvg.info <- hvg.info[order(hvg.info$gene.dispersion.scaled, decreasing = T), ]
  
  topHVGenes <- rownames(hvg.info)[1:2000]
  
  topGenes = topHVGenes
  
  stepwise_reg_obj <- stepwiseRegressionMatrixFS(ggc_original, topGenes, topRangeGenes, num_genes = 1000)
  print(stepwise_reg_obj$st)
  
  list_out <- list()
  num_neighbours <- c("0", "1", "5", "10", "20", "30", "50", "70", "100", "200", "400", "800")
  
  for(n in num_neighbours) {
    list_out[[n]] <- list(num_genes = length(stepwise_reg_obj$neighbour_feature_genes[[n]]), si = SilhouetteBenchmarkPerCTPCA(data = data[stepwise_reg_obj$neighbour_feature_genes[[n]], ], grp = group))
    
  }
  
  return(list("SWReg" = list_out))
}


dubstepR_fn <- function(data, data_name) {
  
  is_10X = ifelse(data_name %in% umi_datasets, T, F)
  
  return(DUBStepR(data, is_10X))
}

trendVar_fn <- function(data, data_name) {
  
  return(trendVarFS(data))
}

#' @objective: to compute the correlation range values for all genes in the gene-gene correlation matrix
#' @input: gene-gene correlation matrix
#' @output: returns a list of p-values, adjusted p-values and correlation ranges for each gene 
getCorrelationRange <- function(correlation_matrix) {
  
  # Sort genes based on their correlation range 
  corr_rank = apply(correlation_matrix,2,sort,decreasing=T)
  
  # Select the second largest correlation for each gene
  rank_2nd_corr = corr_rank[3,]
  
  # Select the smallest correlation for each gene
  min_corr = apply(correlation_matrix,2,min)
  
  # Compute the correlation range
  diff_corr = rank_2nd_corr-(0.75*min_corr)
  # names(diff_corr) <- names(rank_2nd_corr)
  
  # Sort genes in descending order of correlation range
  diff_corr <- sort(diff_corr, decreasing = T)
  
  # Return as list
  return(list("range"=diff_corr))
}



silhouette_fn <- function(data) {
  
}

######################### Clustering ############################################################################

ariByCellType <- function(clustering, ground_truth) {
  for(grp in unique(ground_truth)) {
    
  }
}

# Local FlowSOM function 
flowsom_fn <- function(data, grouping) {
  
  grouping_vec <- as.vector(grouping[[1]])
  
  # Create SCE object from data
  sce = SingleCellExperiment(list(counts = data, logcounts = log(data+1)))
  
  # Read function params
  params <- fromJSON(file = paste0("../parameter_settings/FlowSOM.json"))
  
  # Get k for number of ground truth clusters
  k = length(unique(grouping_vec))
  
  # Obtain clustering results
  clustering_results <- apply_FlowSOM(sce = sce, params = params, k = k)
  
  if(!is.null(clustering_results$error)) {
    return(clustering_results$error)
  }
  
  # Compute ARI
  ari <- tryCatch({
    adjustedRandIndex(clustering_results$cluster, grouping_vec)
  }, error = function(e) {
    e
  })
  
  return(ari)
  
}

# Local PCAHC function 
pcahc_fn <- function(data, grouping) {
  
  grouping_vec <- as.vector(grouping[[1]])
  
  # Create SCE object from data
  sce = SingleCellExperiment(list(counts = data, logcounts = log(data+1)))
  
  # Read function params
  params <- fromJSON(file = paste0("../parameter_settings/PCAHC.json"))
  
  # Get k for number of ground truth clusters
  k = length(unique(grouping_vec))
  
  # Obtain clustering results
  clustering_results <- apply_PCAHC(sce = sce, params = params, k = k)
  
  # Compute ARI
  ari <- tryCatch({
    adjustedRandIndex(clustering_results$cluster, grouping_vec)
  }, error = function(e) {
    0.0
  })
  
  return(ari)
  
}


# Local PCAKMeans function 
pcakmeans_fn <- function(data, grouping) {
  
  grouping_vec <- as.vector(grouping[[1]])
  
  # Create SCE object from data
  sce = SingleCellExperiment(list(counts = data, logcounts = log(data+1)))
  
  # Read function params
  params <- fromJSON(file = paste0("../parameter_settings/PCAKmeans.json"))
  
  # Get k for number of ground truth clusters
  k = length(unique(grouping_vec))
  
  # Return clustering results
  clustering_results <- apply_PCAKmeans(sce = sce, params = params, k = k)
  
  # Compute ARI
  ari <- tryCatch({
    adjustedRandIndex(clustering_results$cluster, grouping_vec)
  }, error = function(e) {
    e
  })
  
  return(ari)
  
}

# Local RtsneKMeans function 
rtsnekmeans_fn <- function(data, grouping) {
  
  grouping_vec <- as.vector(grouping[[1]])
  
  # Create SCE object from data
  sce = SingleCellExperiment(list(counts = data, logcounts = log(data+1)))
  
  # Read function params
  params <- fromJSON(file = paste0("../parameter_settings/RtsneKmeans.json"))
  
  # Get k for number of ground truth clusters
  k = length(unique(grouping_vec))
  
  # Return clustering results
  clustering_results <- apply_RtsneKmeans(sce = sce, params = params, k = k)
  
  if(!is.null(clustering_results$error)) {
    return(clustering_results$error)
  }
  
  # Compute ARI
  ari <- tryCatch({
    adjustedRandIndex(clustering_results$cluster, grouping_vec)
  }, error = function(e) {
    e
  })
  
  return(ari)
  
}

# Local RtsneKMeans function 
seurat_fn <- function(data, grouping) {
  
  grouping_vec <- as.vector(grouping[[1]])
  
  if (is.null(names(grouping_vec))) {
    names(grouping_vec) <- colnames(data)
  }
  # grouping_vec <- unlist(grouping)
  
  # Create SCE object from data
  sce = SingleCellExperiment(list(counts = data, logcounts = log(data+1)))
  
  # Read function params
  params <- fromJSON(file = paste0("../parameter_settings/Seurat.json"))
  
  # Get k for number of ground truth clusters
  k = length(unique(grouping_vec))
  
  # Return clustering results
  res_params <- c(0.4, 0.8, 1.2)
  results_list <- list()
  max_ari = 0.0
  best_clustering <- list()
  for(res in res_params) {
    clustering_results <- apply_Seurat(sce = sce, params = params, resolution = res)
    
    # Compute ARI
    if(any(is.na(clustering_results$cluster)) | (length(clustering_results$cluster) != length(grouping_vec))) {
      ari_list <- list("ari" = 0.0)
    } else {
      ari_list <- tryCatch({
        list("ari" = adjustedRandIndex(clustering_results$cluster, grouping_vec))
      }, error = function(e) {
        list("ari" = 0.0, "error" = e)
      })
    }
    
    if(ari_list[["ari"]] > max_ari) {
      max_ari = ari_list[["ari"]]
      results_list <- ari_list
      best_clustering <- clustering_results
    }
    
  }
  
  # return(append(list("cluster" = best_clustering$cluster, "time" = best_clustering$st), results_list))
  return(ari_list$ari)
}


stepwise_reg_ari_fn <- function(data, group) {
  
  # Remove mito genes
  data <- data[!grepl(pattern = "^MT-", x = rownames(data)), ]
  data <- data[!grepl(pattern = "^MT_", x = rownames(data)), ]
  
  # Remove ERCC genes
  data <- data[!grepl(pattern = "^ERCC-", x = rownames(data)), ]
  data <- data[!grepl(pattern = "^ERCC_", x = rownames(data)), ]
  
  # Remove RPL and RPS genes
  data <- data[!grepl(pattern = "^RPS", x = rownames(data)), ]
  data <- data[!grepl(pattern = "^RPL", x = rownames(data)), ]
  
  # Log transform input data matrix
  log.data <- log(1+data)
  
  # Compute gene-gene correlation matrix
  system.time({
    ggc_original = WGCNA::cor(t(as.matrix(log.data)),method="pearson",nThreads = 8)
  })
  
  # Smooth out diagonal by setting second largest correlation (i.e. largest correlation of a gene with another gene)
  for(i in 1:ncol(ggc_original)) {
    ggc_original[i, i] <- sort(ggc_original[i, ], decreasing = T)[2]
  }
  
  ### Step 1: Choose the top genes based on correlation range
  rangeList <- getCorrelationRange(ggc_original)
  topRangeGenes <- names(rangeList$range)[1:1000]
  
  povList <- getPOV(ggc_original)
  topPOVGenes <- names(povList$POV)[1:1000]
  
  topGenes <- union(topRangeGenes, topPOVGenes)
  
  stepwise_reg_obj <- stepwiseRegressionMatrixFS(ggc_original, topGenes, num_genes = 1000)
  print(stepwise_reg_obj$st)
  
  list_out <- list()
  num_neighbours <- c("0", "5", "10", "20", "30", "50", "100")
  
  ### Cluster cells using the 4 existing clustering methods
  clustering_fns <- c("flowsom_fn", "pcakmeans_fn", "pcahc_fn", "seurat_fn")
  
  for(clust_fn in clustering_fns) {
    print(clust_fn)
    
    list_out[[clust_fn]] <- list()
    
    for(n in num_neighbours) {
      
      list_out[[clust_fn]][[n]] <- get(clust_fn)(data = data[stepwise_reg_obj$neighbour_feature_genes[[n]], ], grouping = group)
    }
  }
  
  return(list("SWReg" = list_out))
  
}
